from serial import Serial as _Serial

from leads.comm import ConnectionBase as _Connection, Service as _Service


class SerialConnection(_Connection):
    def __init__(self, service: _Service, serial: _Serial, port: str, remainder: bytes = b"") -> None:
        super().__init__(service, remainder)
        self._serial: _Serial = serial
        self._port: str = port

    def closed(self) -> bool:
        return self._serial.closed

    def _require_open_serial(self, mandatory: bool = True) -> _Serial:
        if mandatory and self.closed():
            raise IOError("An open serial is required")
        return self._serial

    def receive(self, chunk_size: int = 1) -> bytes | None:
        if self._remainder != b"":
            return self.use_remainder()
        try:
            msg = chunk = b""
            while b";" not in chunk:
                msg += (chunk := self._require_open_serial().read(chunk_size))
            return self.with_remainder(msg)
        except IOError:
            return

    def send(self, msg: bytes) -> None:
        self._require_open_serial().write(msg + b";")
        if msg == b"disconnect":
            self.close()

    def close(self) -> None:
        self._require_open_serial(False).close()



from leads import L, controller, MAIN_CONTROLLER, get_controller, WHEEL_SPEED_CONTROLLER, SRWDataContainer
from leads.config import get_config
from leads_gui import Config
from leads.comm import Callback, Service, ConnectionBase
from leads_arduino import ArduinoMicro
from leads_emulation import SRWRandom

config = get_config(Config)
WHEEL_SPEED_CONTROLLER_PORT: str = config.get("wheel_speed_controller_port", "COM3")
WHEEL_RADIUS: float = config.get("wheel_radius", 159.5)  # 20 inches


@controller(MAIN_CONTROLLER)
class VeCController(SRWRandom):
    def read(self) -> SRWDataContainer:
        return SRWDataContainer(*get_controller(WHEEL_SPEED_CONTROLLER).read())


class WheelSpeedControllerCallback(Callback):
    def on_connect(self, service: Service, connection: ConnectionBase) -> None:
        L.debug("Wheel speed controller connected")

    def on_receive(self, service: Service, msg: bytes) -> None:
        get_controller(WHEEL_SPEED_CONTROLLER).update(float(msg.decode()))

    def on_fail(self, service: Service, error: Exception) -> None:
        L.error(str(error))


@controller(WHEEL_SPEED_CONTROLLER, MAIN_CONTROLLER, (WHEEL_SPEED_CONTROLLER_PORT, WheelSpeedControllerCallback()))
class WheelSpeedController(ArduinoMicro):
    _wheel_speed: float = 0

    def update(self, data: float) -> None:
        self._wheel_speed = data * WHEEL_RADIUS * .0006

    def read(self) -> [float, float]:
        return self._wheel_speed, self._wheel_speed


_ = None  # null export



from RPi import GPIO as _GPIO

from leads import Controller as _Controller


class RaspberryPi4B(_Controller):
    def __init__(self, srw_mode: bool = True) -> None:
        super().__init__()
        self._srw_mode: bool = srw_mode

    def initialize(self, *parent_tags: str) -> None:
        _GPIO.setmode(_GPIO.BCM)
